package axel;

import java.io.*;
import java.util.*;
import util.Cons;
import util.DoubleComparator;
import util.Hypothesis;
import util.Point;

public class AdaBoostClusterer {
    private static float[][] precision;
    private static Map<Integer, List<Hypothesis>> adaBoostMap;
    
    //train AdaBoostClusterer: one movement vs others
    public static void trainAdaBoost() {
        Map<Integer, List<Point>> training_samples = PointStatistics.training_data;
        adaBoostMap = new HashMap<Integer, List<Hypothesis>>();

        int num_samples = 0;
        for (Integer moviment : training_samples.keySet()) {
            num_samples += training_samples.get(moviment).size();
        }

        double[] D = new double[num_samples];

        for (Integer mov_train : training_samples.keySet()) {

            boolean[] y = new boolean[num_samples];
            int i = 0;
            for (Integer mov_dades : training_samples.keySet()) {
                for (Point sample : training_samples.get(mov_dades)) {
                    if (mov_dades == mov_train) {
                        y[i] = true;
                    } else {
                        y[i] = false;
                    }
                    i++;
                }
            }

            for (i = 0; i < D.length; i++) {
                D[i] = (double) 1 / D.length;
            }

            List<Hypothesis> llista_hypothesis = new ArrayList<Hypothesis>();
            int num_features = training_samples.get(mov_train).get(0).size();

            for (int t = 0; t < Cons.numHypothesis_training; t++) {

                Hypothesis hypothesis_error_min = null;

                //per cada feature:
                for (int index_feature = 0; index_feature < num_features; index_feature++) {

                    double[] input = new double[num_samples];
                    int index_sample = 0;
                    for (Integer mov_dades : training_samples.keySet()) {
                        for (Point sample : training_samples.get(mov_dades)) {
                            input[index_sample] = sample.get(index_feature);
                            index_sample++;
                        }
                    }

                    Hypothesis hypothesis = obtainHypothesis_from_one_feature(D, input, y);
                    hypothesis.index_feature = index_feature;

                    if (hypothesis_error_min == null) {
                        hypothesis_error_min = hypothesis;
                    }
                    if (hypothesis_error_min != null && hypothesis.error < hypothesis_error_min.error) {
                        hypothesis_error_min = hypothesis;
                    }
                }

                double[] input = new double[num_samples];
                int index_sample = 0;
                for (Integer mov_dades : training_samples.keySet()) {
                    for (Point sample : training_samples.get(mov_dades)) {
                        input[index_sample] = sample.get(hypothesis_error_min.index_feature);
                        index_sample++;
                    }
                }

                update_D(D, hypothesis_error_min, input, y);

                if (hypothesis_error_min.error >= 0.5) {
                    System.out.println("index_feature: " + hypothesis_error_min.index_feature + " " + Cons.movements[mov_train] + " error>=0.5");
                    break;
                }
                if (hypothesis_error_min.error == 0) {
                    llista_hypothesis.add(hypothesis_error_min);
                    System.out.println("index_feature: " + hypothesis_error_min.index_feature + " " + Cons.movements[mov_train] + " error==0");
                    break;
                }

                llista_hypothesis.add(hypothesis_error_min);
            }

            adaBoostMap.put(mov_train, llista_hypothesis);
        }
    }
    
    public static void saveMapToFile(String fileName){
        try {
            ObjectOutputStream file = new ObjectOutputStream(new FileOutputStream(fileName));
            file.writeObject(adaBoostMap);
            file.close();
        }
        catch (IOException ex) {
            ex.printStackTrace();
        }        
    }
    
    public static void loadModelFromFile(String fileName){
        try {
            ObjectInputStream file = new ObjectInputStream(new FileInputStream(fileName));
            adaBoostMap = (Map<Integer, List<Hypothesis>>) file.readObject();
            file.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    //update weight distribution
    static private void update_D(double[] D, Hypothesis hypothesis, double[] input, boolean y[]) {
        //set thresholds for errors
        boolean greater_threshold = hypothesis.greater_threshold;
        double threshold = hypothesis.threshold;
        //parameter alfa for distribution
        double alfa = hypothesis.alfa;

        double suma_D = 0;
        for (int sample = 0; sample < D.length; sample++) {
            boolean error = false;
            if (y[sample]) {
                if (greater_threshold && input[sample] < threshold) {
                    error = true;
                }
                if (!greater_threshold && input[sample] > threshold) {
                    error = true;
                }
            } else {
                if (greater_threshold && input[sample] > threshold) {
                    error = true;
                }
                if (!greater_threshold && input[sample] < threshold) {
                    error = true;
                }
            }

            if (error) {
                D[sample] *= Math.exp(alfa);
            } else {
                D[sample] *= Math.exp(-alfa);
            }
            suma_D += D[sample];
        }
        for (int i = 0; i < D.length; i++) {
            D[i] /= suma_D;
        }
    }
    
    //get weak hypothesis for one feature (decision stump)
    private static Hypothesis obtainHypothesis_from_one_feature(double[] D, double[] input, boolean[] y) {
        //order values for one feature
        List<Double> input_llista_ordenada = new ArrayList<Double>();
        for (int i = 0; i < input.length; i++) {
            input_llista_ordenada.add(input[i]);
        }
        Collections.sort(input_llista_ordenada);

        double[] input_ordenat = new double[input.length];
        for (int i = 0; i < input.length; i++) {
            input_ordenat[i] = input_llista_ordenada.get(i);
        }

        //set initial parameters
        double error_min = 1;
        double threshold_min = 0;
        boolean greater_threshold_min = false;
        
        boolean greater_threshold = true;
        for (int j = 0; j <= input_ordenat.length; j++) {
            //set threshold
            double threshold = 0;
            if (j == 0) {
                threshold = input_ordenat[j] - 0.05;
            } else {
                if (j == input_ordenat.length) {
                    threshold = input_ordenat[j - 1] + 0.05;
                } else {
                    if (input_ordenat[j] == input_ordenat[j - 1]) {
                        continue;
                    }
                    threshold = (input_ordenat[j] + input_ordenat[j - 1]) / 2;
                }
            }
            //get error for a weak hypothesis
            double error = 0;           
            for (int i = 0; i < input.length; i++) {
                if (y[i]) {
                    if (greater_threshold && input[i] < threshold) {
                        error += D[i];
                    }
                } else {
                    if (greater_threshold && input[i] > threshold) {
                        error += D[i];
                    }
                }
            }

            //find weak hypothesis with min error
            if (error < error_min) {
                error_min = error;
                threshold_min = threshold;
                greater_threshold_min = greater_threshold;
            }
        }

        greater_threshold = false;
        for (int j = 1; j < input_ordenat.length; j++) {
            //set threshold
            double threshold = 0;
            if (j == 0) {
                threshold = input_ordenat[j] - 0.05;
            } else {
                if (j == input_ordenat.length) {
                    threshold = input_ordenat[j - 1] + 0.05;
                } else {
                    if (input_ordenat[j] == input_ordenat[j - 1]) {
                        continue;
                    }
                    threshold = (input_ordenat[j] + input_ordenat[j - 1]) / 2;
                }
            }
            //get error for a weak hypothesis
            double error = 0;
            for (int i = 0; i < input.length; i++) {
                if (y[i]) {                   
                    if (!greater_threshold && input[i] > threshold) {
                        error += D[i];
                    }
                } else {
                    if (!greater_threshold && input[i] < threshold) {
                        error += D[i];
                    }
                }
            }

            //find weak hypothesis with min error
            if (error < error_min) {
                error_min = error;
                threshold_min = threshold;
                greater_threshold_min = greater_threshold;
            }
        }
        //create weak hypothesis with min error
        Hypothesis hypothesis = new Hypothesis();

        hypothesis.threshold = threshold_min;
        hypothesis.greater_threshold = greater_threshold_min;
        hypothesis.error = error_min;

        if (error_min < 0.000001) {
            error_min = 0.000001;
        }
        
        hypothesis.alfa = 0.5 * Math.log((1 - error_min) / error_min);

        return hypothesis;
    }
    
    public static String testAdaBoost(String subname) {

        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_adaBoost_" + subname + "_.txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }

        precision = new float[Cons.movements.length][Cons.movements.length];

        int undefined_mov = Cons.movements.length - 1;
        int allSamples = 0;
        int errorCounter = 0;
        int undefinedCounter = 0;

        int i = 0;
        for (Point sample : PointStatistics.test_data) {

            int mov_real = PointStatistics.test_movements_sequence.get(i);

            allSamples++;
            int prediction = prediction_by_single_list_hypothesis(sample, file);

            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);

            if (prediction != mov_real && prediction != undefined_mov) {
                errorCounter++;
                file.println("error n?: " + errorCounter);
            }
            if (prediction == undefined_mov) {
                undefinedCounter++;
            }
            precision[mov_real][prediction] += 1;
            i++;
        }

        double accuracy = (double) (allSamples-errorCounter) / allSamples;
        double undefined_rate = (double) undefinedCounter / allSamples;
        file.println("***************************************");
        file.println("accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")");
        file.println("undefined rate:  " + undefined_rate + " (" + undefinedCounter + "/" + allSamples + ")");
        file.println("***************************************");
        for (i = 0; i < precision.length; i++) {
            int num_predictions = 0;
            for (int j = 0; j < precision[0].length; j++) {
                num_predictions += precision[i][j];
            }
            file.println(" ");
            file.print(Cons.movements[i]);
            for (int j = 0; j < precision[0].length; j++) {
                String result = new Float(precision[i][j] / num_predictions).toString();
                if (result.length() < 4) {
                    result = result.concat("0");
                } else {
                    result = result.substring(0, 4);
                }
                file.print(result + " ");
            }
        }
        if (file != null) {
            file.close();
        }
        return "accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")";
    }
    
    public static int prediction_by_single_list_hypothesis(Point sample, PrintWriter file) {
        Map<Integer, Double> mov_hypothesis_map = new HashMap<Integer, Double>();

        for (Integer mov : adaBoostMap.keySet()) {

            double hypothesis_value = 0;
            double suma_alfa = 0;
            int hypothesis_counter = Cons.numHypothesis_tracking;

            List<Hypothesis> llista_hypothesis = adaBoostMap.get(mov);

            if (llista_hypothesis.isEmpty()) {
                continue;
            }

            for (Hypothesis hypothesis : llista_hypothesis) {

                boolean greater_threshold = hypothesis.greater_threshold;
                double threshold = hypothesis.threshold;
                double alfa = hypothesis.alfa;

                boolean positive_y = false;

                double input = sample.get(hypothesis.index_feature);

                if (greater_threshold && input > threshold) {
                    positive_y = true;
                }
                if (!greater_threshold && input < threshold) {
                    positive_y = true;
                }

                if (positive_y) {
                    hypothesis_value += alfa;
                } else {
                    hypothesis_value -= alfa;
                }

                suma_alfa += alfa;

                if (--hypothesis_counter == 0) {
                    break;
                }
            }

            mov_hypothesis_map.put(mov, ((double) hypothesis_value / suma_alfa));
        }
        
        Map<Integer, Double> mov_reliability = mov_hypothesis_map;

        DoubleComparator comparator = new DoubleComparator(mov_reliability);
        TreeMap<Integer, Double> ordered_mov_reliability = new TreeMap<Integer, Double>(comparator);
        ordered_mov_reliability.putAll(mov_reliability);

        int best_mov = ordered_mov_reliability.firstKey();
        int second_mov = best_mov;

        if (file != null) {
            file.println("-------------------------");
        }
        int position = 0;
        for (Integer mov : ordered_mov_reliability.keySet()) {
            if (++position == 2) {
                second_mov = mov;
            }
            if (file != null) {
                file.println(Cons.movements[mov] + ": " + mov_reliability.get(mov));
            }
        }
        if (file != null) {
            file.println("-------------------------");
        }

        int amplifier = 1;

        if (mov_reliability.get(best_mov) > 0 && mov_reliability.get(best_mov) - mov_reliability.get(second_mov) <= Cons.undefinedMarginPositive * amplifier) {
            best_mov = Cons.UNDEFINED;
        } else if (mov_reliability.get(best_mov) < 0 && mov_reliability.get(best_mov) - mov_reliability.get(second_mov) <= Cons.undefinedMarginNegative * amplifier) {
            best_mov = Cons.UNDEFINED;
        }

        return best_mov;
    }
}