package axel;

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */


import Jama.Matrix;
import java.io.*;
import java.util.*;
import util.Cons;
import util.DoubleComparator;
import util.Point;
import util.Util;


/**
 *
 * @author Dunst
 */
public class BoostedKMeansClusterer {    
    private static Map<Integer, Point> centroids = new HashMap<Integer, Point>();
    private static float[][] precision;
    
    public static void trainBoostedKMeans(){                
        Map<Integer, List<Point>> training_samples = PointStatistics.training_data;
        Map<Integer, List<Point>> weighted_training_samples = PointStatistics.training_data;        
        List<Double> CQ = new ArrayList<Double>();
        
        //initialize algorithm
        int num_samples = 0;
        for (Integer moviment : training_samples.keySet()) {
            num_samples += training_samples.get(moviment).size();
        }
        
        double[] D = new double[num_samples];
        int i = 0;
        for(Integer moviment : training_samples.keySet()){
            for (Point sample: training_samples.get(moviment)){
                D[i] = (double)1/training_samples.get(moviment).size();
                i++;
            }
        }
        weighted_training_samples = addWeight(training_samples, D);
        
        for (int t = 0; t < Cons.numHypothesis_training; t++) {            
            
            for (Integer mov: weighted_training_samples.keySet()){
                Point c = calculateCentroid(weighted_training_samples.get(mov));
                centroids.put(mov, c);
            }
            
            Matrix h = new Matrix(weighted_training_samples.keySet().size(), num_samples);
            h = calculateH(weighted_training_samples, num_samples);
            
            CQ = calculateCQ(h, num_samples);
            
            double eloss = calculatePseudoLoss(D, CQ);
            
            double b = Math.log((1-eloss)/eloss);
            
            D = updateD(D, b, CQ);
            weighted_training_samples = addWeight(training_samples, D);
        }
            
    }
    
    public static void saveClustersToFile(String fileName){
        try {
            ObjectOutputStream file = new ObjectOutputStream(new FileOutputStream(Cons.root_dir+"/"+fileName+".txt"));
            file.writeObject(centroids);
            file.close();
        }
        catch (IOException ex) {
            ex.printStackTrace();
        }        
    }
    
    public static void loadClustersFromFile(String fileName){
        try {
            ObjectInputStream file = new ObjectInputStream(new FileInputStream(Cons.root_dir+"/"+fileName+".txt"));
            centroids = (Map<Integer, Point>) file.readObject();
            file.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    private static Matrix calculateH(Map<Integer, List<Point>> weighted_training_samples, int num_samples){
        Matrix h = new Matrix(weighted_training_samples.keySet().size(), num_samples);
        List<Point> list_samples = new ArrayList<Point>();
        for (Integer mov: weighted_training_samples.keySet()){
            list_samples.addAll(weighted_training_samples.get(mov));
        }
        
        for (int i=0; i<list_samples.size(); i++){
            int j = 0;
            double d1 = getD1(weighted_training_samples.keySet(), list_samples.get(i));
            for (Integer mov: weighted_training_samples.keySet()){
                double d0 = Util.euclidDistance(centroids.get(mov), list_samples.get(i));
                h.set(j, i, d1/d0);
                j++;
            }
        }
        
        double s = 0;
        for (int j=0; j<weighted_training_samples.keySet().size();j++)
            s = s+h.get(j, 0);
        return h;
    }
    
    private static double getD1(Set<Integer> keys, Point sample){
        double d1 = 0;
        
        for (Integer mov: keys){
            d1 += Util.euclidDistance(sample, centroids.get(mov));//  calculateEuclidDistance(sample, centroids.get(mov));
        }
        return d1;
    }
    
    private static double calculatePseudoLoss(double[] D, List<Double> CQ){        
        
        double loss = 0;
        for (int i=0; i<D.length; i++){
            if (!Double.isNaN(CQ.get(i))){
                loss += D[i]*CQ.get(i);
            }
        }   

        return loss/2;       
    }
    
    //entropy-based measure
    private static List<Double> calculateCQ(Matrix h, int numberOfSamples){
        List<Double> CQ = new ArrayList<Double>();
        CQ = fillWith(CQ, 0, numberOfSamples);        
        for (int i=0; i<numberOfSamples; i++){
            double cq = 0;
            
            for (int j=0; j<h.getRowDimension(); j++){
                cq = h.get(j, i)*Math.log(h.get(j, i));
            }
            CQ.set(i, -cq);
        }
        return CQ;
    }
    
    
    private static Point calculateCentroid(List<Point> input){
        List<Float> centroid = new ArrayList<Float>();
        centroid = fillWith(centroid, 0, input.get(0).size());
        
        for (Point sample: input){
            for (int i=0; i<sample.size(); i++){
                centroid.set(i, centroid.get(i)+ sample.get(i));
            }
        }
        
        for (int i=0; i<centroid.size(); i++){
            centroid.set(i, centroid.get(i)/input.size());
        }
        
        return new Point(centroid);
    }

    private static List<Float> fillWith(List<Float> x, float a, int size){
        for (int i=0; i<size;i++)
            x.add(a);
        return x;
    }
    
    private static List<Double> fillWith(List<Double> x, double a, int size){
        for (int i=0; i<size;i++)
            x.add(a);
        return x;
    }
    
    private static double calculateEuclidDistance(Point a, Point b){
        double d = 0;
        
        for (int i=0; i< a.size(); i++){
            d = (a.get(i)-b.get(i))*(a.get(i)-b.get(i));
        }
        if (d != 0)
            d = Math.sqrt(d);
        
        return d;
    }
    
    public static Map<Integer, List<Point>> addWeight(Map<Integer, List<Point>> samples, double[] D){
        Map<Integer, List<Point>> weighted_training_samples = new HashMap<Integer, List<Point>>();
        int i= 0;
        for (Integer mov: samples.keySet()){
            List<Point> w_samples = new ArrayList<Point>();            
            for (Point sample: samples.get(mov)){
                sample.setWeigth(D[i]);
                w_samples.add(sample);
                i++;
            }
            weighted_training_samples.put(mov, w_samples);            
        }
        
        return weighted_training_samples;
    }
    
    public static double[] updateD(double[] D, double b, List<Double> CQ){
        double[] updatedD = new double[D.length];
        double sum = 0;
        
        for (int i=0; i<updatedD.length; i++){
            updatedD[i] = D[i]*Math.pow(b, CQ.get(i));
            sum += updatedD[i];
        }
        for (int i=0; i<updatedD.length; i++){
            updatedD[i] = updatedD[i]/sum;
        }
        
        return updatedD;
    }
    
    public static void testBoostedKMeans(String subname){
        int errorCounter = 0;
        precision = new float[Cons.movements.length][Cons.movements.length];
        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_boostedKmeans_" + subname + ".txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        
        for (int i = 0; i< PointStatistics.test_data.size(); i++){
            int mov_real = PointStatistics.test_movements_sequence.get(i);
            
            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            int prediction = prediction(PointStatistics.test_data.get(i), file);
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);
            
            if (prediction != mov_real && prediction != Cons.UNDEFINED) {
                errorCounter++;               
            }
            precision[mov_real][prediction] += 1;
        }
        double allSamples = PointStatistics.test_data.size();
        double accuracy = (allSamples-errorCounter) / allSamples;
        file.println("***************************************");
        file.println("accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")");
        file.println("***************************************");
        for (int i = 0; i < precision.length; i++) {
            int num_predictions = 0;
            for (int j = 0; j < precision[0].length; j++) {
                num_predictions += precision[i][j];
            }
            file.println(" ");
            file.print(Cons.movements[i]);
            for (int j = 0; j < precision[0].length; j++) {
                String result = new Float(precision[i][j] / num_predictions).toString();
                if (result.length() < 4) {
                    result = result.concat("0");
                } else {
                    result = result.substring(0, 4);
                }
                file.print(result + " ");
            }
        }
        if (file != null) {
            file.close();
        }
    }
    
    private static int prediction(Point sample, PrintWriter file){
        Map<Integer, Double> mov_affinity = new HashMap<Integer, Double>();
        for (Integer mov: centroids.keySet()){
            double d = calculateEuclidDistance(sample, centroids.get(mov));             
            double d1 = getD1(centroids.keySet(), sample);
            mov_affinity.put(mov, d/d1);
            file.println(mov+": "+mov_affinity.get(mov));
        }
        
        DoubleComparator comparator = new DoubleComparator(mov_affinity);
        TreeMap<Integer, Double> ordered_mov_affinity = new TreeMap<Integer, Double>(comparator);
        ordered_mov_affinity.putAll(mov_affinity);

        int best_mov = ordered_mov_affinity.lastKey();
        
        return  best_mov;
    }
}
