/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package axel;

import java.io.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import util.Cons;
import util.Point;

/**
 *
 * @author Dunst
 */
public class GMMEMClusterer {
    private static Map<Integer, List<Point>> training_samples;
    private static Map<Integer, GaussianMixtureModel> gmmModel;
    private static float[][] precision;
    
    public static void trainGMM(int numOfGaussian){
        training_samples = PointStatistics.training_data;
        gmmModel = new HashMap<Integer, GaussianMixtureModel>();
        for (Integer mov: training_samples.keySet()){
            GaussianMixtureModel gmm = new GaussianMixtureModel(mov, training_samples.get(mov), numOfGaussian);
            GaussianMixtureModel gmm_curr = null;
            double logPrMax = 0;
            //divide samples set into clasters and find centroids for them by running Kmeans
            //initilazize weigth component for each cluster to 1/number_of_samples_in_set
            //calculate means and covariance matrix for samples set
            gmm.init();            
            //start iteration
            for (int t = 0; t < Cons.numHypothesis_training; t++){
                gmm.runClusterer();
//                double logPr = gmm.getLogLikelihood();
//                if (Math.abs(logPrMax) < Math.abs(logPr)){
//                    logPrMax = logPr;
                    gmm_curr = gmm;
//                }
            }
            gmmModel.put(mov, gmm_curr);
        }
    }
    
    public static void saveModelToFile(String fileName){
        try {
            ObjectOutputStream file = new ObjectOutputStream(new FileOutputStream(Cons.root_dir+"/"+fileName+".txt"));
            file.writeObject(gmmModel);
            file.close();
        }
        catch (IOException ex) {
            ex.printStackTrace();
        }        
    }
    
    public static void loadModelFromFile(String fileName){
        try {
            ObjectInputStream file = new ObjectInputStream(new FileInputStream(Cons.root_dir+"/"+fileName+".txt"));
            gmmModel = (Map<Integer, GaussianMixtureModel>) file.readObject();
            file.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    public static void testGMM(String subname){
        int errorCounter = 0;
        precision = new float[Cons.movements.length][Cons.movements.length];
        
        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_GMMClusterer_" + subname + ".txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        for (int i = 0; i< PointStatistics.test_data.size(); i++){
            Point sample = PointStatistics.test_data.get(i);
            int mov_real = PointStatistics.test_movements_sequence.get(i);
            int prediction = prediction(sample);
            
            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);
            file.println("\n-------------------------");
            
            if (prediction != mov_real ) {
                errorCounter++;
            }
            precision[mov_real][prediction] += 1;
        }
        double allSamples = PointStatistics.test_data.size();
        double accuracy = (allSamples-errorCounter) / allSamples;
        file.println("***************************************");
        file.println("accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")");
        file.println("***************************************");
        for (int i = 0; i < precision.length; i++) {
            int num_predictions = 0;
            for (int j = 0; j < precision[0].length; j++) {
                num_predictions += precision[i][j];
            }
            file.println(" ");
            file.print(Cons.movements[i]);
            for (int j = 0; j < precision[0].length; j++) {
                String result = new Float(precision[i][j] / num_predictions).toString();
                if (result.length() < 4) {
                    result = result.concat("0");
                } else {
                    result = result.substring(0, 4);
                }
                file.print(result + " ");
            }
        }
        
        if (file != null) {
            file.close();
        }        
    }
    
    private static int prediction(Point sample){
        int prediction = -1;
        double logPrMax = 0;
        for (Integer mov: gmmModel.keySet()){
            GaussianMixtureModel cluster = gmmModel.get(mov);
            double logPr = cluster.getPdf(sample);
            if (logPr>logPrMax){
                prediction = cluster.getId();
                logPrMax = logPr;
            }
        }
        
        return prediction;        
    }
}
 