package axel;


import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import util.Cluster;
import util.Cons;
import util.Point;

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

/**
 *
 * @author Dunst
 */
public class KMeansClusterer {
    private static Map<Integer, Cluster> clusters = new HashMap<Integer, Cluster>();
    private static float[][] precision;
    
    public static void computeCentroidsForTrainData(){
        //calculate init clusters
        for(Integer mov: PointStatistics.training_data.keySet()){
            Cluster k = new Cluster(mov);
            for(Point sample: PointStatistics.training_data.get(mov)){
                k.add(sample);
            }
            k.resetCenter();
            clusters.put(mov, k);
        }
    }
    
    public static void saveClustersToFile(String fileName){
        try {
            ObjectOutputStream file = new ObjectOutputStream(new FileOutputStream(Cons.root_dir+"/"+fileName+".txt"));
            file.writeObject(clusters);
            file.close();
        }
        catch (IOException ex) {
            ex.printStackTrace();
        }        
    }
    
    public static void loadClustersFromFile(String fileName){
        try {
            ObjectInputStream file = new ObjectInputStream(new FileInputStream(Cons.root_dir+"/"+fileName+".txt"));
            clusters = (Map<Integer, Cluster>) file.readObject();
            file.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    public static void testKmeans(String subname){
        int errorCounter = 0;
        precision = new float[Cons.movements.length][Cons.movements.length];
        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_KMeansClusterer_" + subname + ".txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        for (int i = 0; i< PointStatistics.test_data.size(); i++){
            Point sample = PointStatistics.test_data.get(i);
            int mov_real = PointStatistics.test_movements_sequence.get(i);
            int prediction = findNearestCluster(sample, clusters);
            
            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);
            file.println("\n-------------------------");
            
            if (prediction != mov_real ) {
                errorCounter++;
            }
            precision[mov_real][prediction] += 1;
        }
        double allSamples = PointStatistics.test_data.size();
        double accuracy = (allSamples-errorCounter) / allSamples;
        file.println("***************************************");
        file.println("accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")");
        file.println("***************************************");
        for (int i = 0; i < precision.length; i++) {
            int num_predictions = 0;
            for (int j = 0; j < precision[0].length; j++) {
                num_predictions += precision[i][j];
            }
            file.println(" ");
            file.print(Cons.movements[i]);
            for (int j = 0; j < precision[0].length; j++) {
                String result = new Float(precision[i][j] / num_predictions).toString();
                if (result.length() < 4) {
                    result = result.concat("0");
                } else {
                    result = result.substring(0, 4);
                }
                file.print(result + " ");
            }
        }
        
        if (file != null) {
            file.close();
        }        
    }
    
    public static void testKMeansOnAllData(String subname){
        List<Point> samples = new ArrayList<Point>();        
        List<Point> good_samples = new ArrayList<Point>();        
        
        //calculate init clusters
        for(Integer mov: PointStatistics.training_data.keySet()){
            Cluster k = new Cluster(mov);
            for(Point sample: PointStatistics.training_data.get(mov)){
                samples.add(sample);
                k.add(sample);
            }
            k.resetCenter();
            clusters.put(mov, k);
        }
        
        int train_n = samples.size();
        //add test samples
        samples.addAll(PointStatistics.test_data);
        
        //set weights
        for(Point p: samples){
            p.setWeigth((double)1/samples.size());
        }
        double eloss_min = Double.MAX_VALUE;
        for (int t=0; t<Cons.numHypothesis_training; t++){
            //run iteration
            
            //get membership degree for point to center of each class
            Map<Integer, List<Double>> d = new HashMap<Integer, List<Double>>();        
            for (Integer mov: clusters.keySet()){
                Cluster cluster = clusters.get(mov);
                List<Double> d_mov_i = new ArrayList<Double>();
                for (Point sample: samples){
                    double d0 = cluster.euclidDistance(sample);
                    d0 = d0/sumDClusters(sample, clusters);                    
                    d_mov_i.add(d0);
                }
                d.put(cluster.getId(), d_mov_i);
            }

            //reassign labels
            int i= 0;
            for (Point sample: samples){
                sample.setClusterId(findNearestCluster(sample, clusters));
            }
            //rebuild clusters
            for (Integer mov: clusters.keySet()){
                Cluster k = clusters.get(mov);
                k.clear();
                for (Point sample: samples){
                    if (k.getId() == sample.getClusterId())
                        k.add(sample);
                }
                k.resetCenter();
            }
            
            //get entropy-based measure for each class
            List<Double> CQ = calculateCQ(d, samples.size()); 
            //get error for classification
            double eloss = calculatePseudoLoss(samples, CQ);
            if (eloss < 0)
                break;
            if (eloss<eloss_min){
                eloss_min = eloss;
                good_samples.clear();
                good_samples.addAll(samples);
            }
            double b = (1-eloss)/eloss;
            
            //update weights
            samples = updateD(samples, b, CQ);
        }
        
        int errorCounter = 0;
        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_KMeansClustererAll_" + subname +".txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        
        int samples_number = PointStatistics.test_data.size();
        for (int i = 0; i< samples_number; i++){
            int mov_real = PointStatistics.test_movements_sequence.get(i);
            int prediction = good_samples.get(train_n+i).getClusterId();
            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);
            file.println("\n-------------------------");
            
            if (prediction != mov_real ) {
                errorCounter++;
            }
        }
        file.println("***************************************");
        file.println("error rate:  " + (double)errorCounter/samples_number+"("+errorCounter+"/"+samples_number+")");
        file.println("***************************************");
        
        if (file != null) {
            file.close();
        }
    }    
    
    private static List<Point> updateD(List<Point> samples, double b, List<Double> CQ){
        double sum = 0;
        
        int i = 0;
        for (Point sample: samples){
            double d = sample.getWeight()*Math.pow(b, CQ.get(i));
            sample.setWeigth(d);
            sum = sum+d;
        }
        
        for (Point sample: samples){
            sample.setWeigth(sample.getWeight()/sum);
        }
        return samples;
    }

    private static int findNearestCluster(Point sample, Map<Integer, Cluster> clusters){
        Cluster nearestCluster = null;
        double nearestDistance = Double.MAX_VALUE;
        for(Integer mov: clusters.keySet()){
            Cluster cluster = clusters.get(mov);
            double distance = cluster.euclidDistance(sample);
            distance = distance/sumDClusters(sample, clusters);
            if (distance < nearestDistance || nearestCluster == null) {
                nearestCluster = cluster;
                nearestDistance = distance;
            }
        }
        return nearestCluster.getId();
    }
    
    //entropy-based measure
    private static List<Double> calculateCQ(Map<Integer, List<Double>> d, int numberOfSamples){
        List<Double> CQ = new ArrayList<Double>();
        
        for (int i=0; i<numberOfSamples; i++){
            double cq = 0;
            
            for (Integer mov: d.keySet()){
                cq += d.get(mov).get(i)*Math.log(d.get(mov).get(i));
            }
            CQ.add(-cq);
        }
        return CQ;
    }
    
    private static double calculatePseudoLoss(List<Point> samples, List<Double> CQ){
        
        double loss = 0;
        for (int i=0; i<samples.size(); i++){
            if (!Double.isNaN(CQ.get(i))){
                loss += samples.get(i).getWeight() *CQ.get(i);
            }
        }   

        return loss/2;       
    }
    
    private static double sumDClusters(Point sample, Map<Integer, Cluster> clusters){
        double sum = 0;
        for (Integer mov: clusters.keySet()){
            Cluster cluster = clusters.get(mov);
            sum += cluster.euclidDistance(sample);
        }
        return sum;
    }
}
