/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package axel;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.*;
import util.Cons;
import util.DoubleComparator;
import util.Point;
import util.Util;

/**
 *
 * @author Dunst
 */
public class KNNClusterer {
    private static float[][] precision;
    private List<Point> trainData = new ArrayList<Point>();
    private int K;
    
    public KNNClusterer(int K){
        this.K = K;
        for(Integer mov: PointStatistics.training_data.keySet()){
            for (Point p: PointStatistics.training_data.get(mov)){
                trainData.add(p);
            }
        }
    }
    
    public KNNClusterer(){
        this(7);
    }
    
    public void test(String subname){
        int errorCounter = 0;
        precision = new float[Cons.movements.length][Cons.movements.length];
        PrintWriter file = null;
        try {
            file = new PrintWriter(new FileWriter(Cons.root_dir + "/sortida_KNNClusterer_" + subname +".txt"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }        
        
        for (int i = 0; i< PointStatistics.test_data.size(); i++){
            int mov_real = PointStatistics.test_movements_sequence.get(i);
            //int prediction = predict(PointStatistics.test_data.get(i));
            int prediction = predictByDistance(PointStatistics.test_data.get(i));
            file.println("window " + PointStatistics.test_mov_window_sequence.get(i) + " del moviment");
            file.println("\n-------------------------");
            file.println("Actual movement: " + Cons.movements[mov_real]);
            file.println("Predicted movem: " + Cons.movements[prediction]);
            file.println("\n-------------------------");
            
            if (prediction != mov_real ) {
                errorCounter++;
            }
            precision[mov_real][prediction] += 1;
        }
        double allSamples = PointStatistics.test_data.size();
        double accuracy = (allSamples-errorCounter) / allSamples;
        file.println("***************************************");
        file.println("accuracy:  " + accuracy + " (" + (allSamples-errorCounter) + "/" + allSamples + ")");
        file.println("***************************************");
        for (int i = 0; i < precision.length; i++) {
            int num_predictions = 0;
            for (int j = 0; j < precision[0].length; j++) {
                num_predictions += precision[i][j];
            }
            file.println(" ");
            file.print(Cons.movements[i]);
            for (int j = 0; j < precision[0].length; j++) {
                String result = new Float(precision[i][j] / num_predictions).toString();
                if (result.length() < 4) {
                    result = result.concat("0");
                } else {
                    result = result.substring(0, 4);
                }
                file.print(result + " ");
            }
        }
        
        if (file != null) {
            file.close();
        }
        
        
    }
    
    public int predict(Point sample){        
        List<Integer> predictions = new ArrayList<Integer>();
        Map<Integer, Double> nearest = new HashMap<Integer, Double>();
        
        for (int i=0; i<sample.size(); i++){
            predictions.add(predictionForOneFeature(i, sample.get(i)));
        }
        
        for (int i=0; i<sample.size(); i++){
            int mov = predictions.get(i);
            if (nearest.containsKey(mov)){
                double n = nearest.get(mov);
                nearest.put(mov, n+1);
            }
            else {
                nearest.put(mov, (double)1);
            }
        }
        
        DoubleComparator comparator = new DoubleComparator(nearest);
        TreeMap<Integer, Double> ordered_nearest = new TreeMap<Integer, Double>(comparator);
        ordered_nearest.putAll(nearest);
        
        return ordered_nearest.firstKey();
    }
    
    public int predictByDistance(Point sample){
        Double[][] distances = new Double[trainData.size()][2];
        Map<Integer, Double> nearest = new HashMap<Integer, Double>();
        
        for (int i=0; i<trainData.size(); i++){
            distances[i][0] = new Double(trainData.get(i).getClusterId());
            distances[i][1] = Util.euclidDistance(sample, trainData.get(i));
        }
        
        Arrays.sort(distances, new Comparator<Double[]>(){
            @Override
            public int compare(Double[] o1, Double[] o2) {
                return o1[1].compareTo(o2[1]);
            }
        });
        
        for (int k = 0; k<K; k++){
            int mov = distances[k][0].intValue();
            if (nearest.containsKey(mov)){
                double n = nearest.get(mov);
                nearest.put(mov, n+1);
            }
            else {
                nearest.put(mov, (double)1);
            }
        }
        
        DoubleComparator comparator = new DoubleComparator(nearest);
        TreeMap<Integer, Double> ordered_nearest = new TreeMap<Integer, Double>(comparator);
        ordered_nearest.putAll(nearest);
        int prediction = ordered_nearest.firstKey();
        
        return prediction;
    }

    private int predictionForOneFeature(int numFeature, float feature){
        int prediction = -1;
        Map<Integer, Double> nearest = new HashMap<Integer, Double>();
        
        Double[][] distances = new Double[trainData.size()][2];
       
        int i = 0;
        for (Point p: trainData){
            distances[i][0] = new Double(p.getClusterId());
            distances[i][1] = calculateDistance(feature, p.get(numFeature));
            i = i+1;
        }
        
        Arrays.sort(distances, new Comparator<Double[]>(){
            @Override
            public int compare(Double[] o1, Double[] o2) {
                return o1[1].compareTo(o2[1]);
            }
        });
        
        for (int k = 0; k<K; k++){
            int mov = distances[k][0].intValue();
            if (nearest.containsKey(mov)){
                double n = nearest.get(mov);
                nearest.put(mov, n+1);
            }
            else {
                nearest.put(mov, (double)1);
            }
        }
        
        DoubleComparator comparator = new DoubleComparator(nearest);
        TreeMap<Integer, Double> ordered_nearest = new TreeMap<Integer, Double>(comparator);
        ordered_nearest.putAll(nearest);
        prediction = ordered_nearest.firstKey();
        
        return prediction;
    }
    
    private Double calculateDistance(float testFeature, float trainFeature){
        return new Double(Math.abs(testFeature - trainFeature));
    }
}
