package axel;


import java.io.*;
import java.util.*;
import util.*;

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

/**
 *
 * @author Dunst
 */
public class PointStatistics {
    public static float[] x, y, z, m;
    // 50 mostres/segon: 128 mostres (2'5 segons aproximadament)
    public static int window = 128;
    public static int half_window = window/2;
    public static int fourth_window = window/4;
    public static int rang = 32;//32;
    
    public static Map<Integer, List<Point>> training_data;
    public static List<Point> test_data;
    public static List<Integer> test_movements_sequence;
    public static List<Integer> test_mov_window_sequence;
    
    private static boolean useFilter;
    private static boolean testPhase;
    
    public  PointStatistics(){
        training_data = new HashMap<Integer, List<Point>>();
        test_data = new ArrayList<Point>();
        test_movements_sequence = new ArrayList<Integer>();
        test_mov_window_sequence = new ArrayList<Integer>();
        
        useFilter = false;
        testPhase = false;
    }
    
    public static void setTest(boolean isTest){
        testPhase = isTest;
    }
    
    public static void setUseFilter(boolean useFilterForData){
        useFilter = useFilterForData;
    }
    
    public static void loadDataFromDir(String directory){
        if (testPhase){
            test_data.clear();
            test_movements_sequence.clear();
            test_mov_window_sequence.clear();
        }
        else{
            training_data.clear();
        }
        File dir = new File(directory);
        String[] llista_files = dir.list();
        for(String file_name : llista_files){
            int moviment = new Integer(file_name.substring(4,5)).intValue();
            loadFile(directory+"/"+file_name, file_name);
            obtainStatistics(moviment);
        }
    }
    
    public static void loadFile(String fullName, String fileName){
        BufferedReader file=null;
//        PrintWriter wfile=null;
        KalmanFilter kf_x = new KalmanFilter(5, 5);
        KalmanFilter kf_y = new KalmanFilter(5, 5);
        KalmanFilter kf_z = new KalmanFilter(5, 5);
        
        try{            
            file = new BufferedReader(new FileReader(fullName));
//            wfile = new PrintWriter(new FileWriter("datafiles/x_fall"+fileName+".txt"));
            //load data from file
            String line;
            List<Float> listX = new ArrayList<Float>();
            List<Float> listY = new ArrayList<Float>();
            List<Float> listZ = new ArrayList<Float>();
            
            int size = 0;
            while((line = file.readLine())!=null){
                size++;
                String[] figures = line.split(" ");
                listX.add(new Float(figures[1]));                
                listY.add(new Float(figures[2]));
                listZ.add(new Float(figures[3]));
            }
            
            float[] x_ = new float[size];
            float[] y_ = new float[size];
            float[] z_ = new float[size];
            float[] m_ = new float[size];
            
            for (int i=0; i<size; i++){                
                x_[i] = listX.get(i);
                y_[i] = listY.get(i);
                //remove the force of gravity
                z_[i] = listZ.get(i) - (float)9.8;                
                m_[i] = (float) Math.sqrt(x_[i]*x_[i]+y_[i]*y_[i]+z_[i]*z_[i]);
                
//                wfile.println(x_[i]);
            }
           
            int inici=0, fi=size;
            boolean set_inici=false;
            float[] m_rang = new float[rang];

            //find low energy part
            for(int i=0; i<size-rang; i+=rang){
                for(int j=0; j<rang; j++){
                    m_rang[j] = m_[i+j];
                }
                float energy = energy(m_rang);

                if(energy>Cons.min_energy_rang && !set_inici){
                    inici = i;
                    set_inici = true;
                }
                if(energy<Cons.min_energy_rang && set_inici){
                    fi = i;
                    break;
                }
            }

            //remove low energy part
            int activity_size = fi-inici;
            if(activity_size<window)
                size = window;
            else{
                size = (activity_size/window)*window;
                fi = inici+size;
            }

            x = new float[size];
            y = new float[size];
            z = new float[size];
            m = new float[size];
            
            if (useFilter){
                kf_x.setInitState(x_[inici], 0.1);
                kf_y.setInitState(y_[inici], 0.1);
                kf_z.setInitState(z_[inici], 0.1);
            }

            for(int i=inici; i<fi; i++){
                if (useFilter){
                    kf_x.correct(x_[i]);
                    kf_y.correct(y_[i]);
                    kf_z.correct(z_[i]);
                    
                    x[i-inici] = (float)kf_x.getState();
                    y[i-inici] = (float)kf_y.getState();
                    z[i-inici] = (float)kf_z.getState();
                    m[i-inici] = (float) Math.sqrt(x[i-inici]*x[i-inici]+y[i-inici]*y[i-inici]+z[i-inici]*z[i-inici]);
                }
                else {
                    x[i-inici] = x_[i];
                    y[i-inici] = y_[i];
                    z[i-inici] = z_[i];
                    m[i-inici] = m_[i];                    
                }
            }            
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        try {
            if(file!=null)
                file.close();
//            if(wfile!=null)
//                wfile.close();
        } catch (IOException ex1) {
            ex1.printStackTrace();
        }        
    }
    
    private static void obtainStatistics(int movement){
        if(!testPhase){
            if(!training_data.containsKey(movement))
                training_data.put(movement, new ArrayList<Point>());
        }
            
        float[] x_w = new float[window];
        float[] y_w = new float[window];
        float[] z_w = new float[window];
        float[] m_w = new float[window];

        //es resta 3 per 1/4, 1 per 1/2 amb aixo la ultima finestra es justament
        //el final de les dades:
        int num_windows = x.length/fourth_window -3;

        //
        for(int w=0; w<num_windows; w++){
            for(int i=0; i<window; i++){
                x_w[i] = x[fourth_window*w+i];
                y_w[i] = y[fourth_window*w+i];
                z_w[i] = z[fourth_window*w+i];
                m_w[i] = m[fourth_window*w+i];
            }

            Point sample = null;

            if(!isNull(m_w))
                 sample = sample(x_w, y_w, z_w, m_w, w);

            if(sample==null)
                continue;            

            if(!testPhase){
                sample.setClusterId(movement);
                training_data.get(movement).add(sample);
            }
            if(testPhase){
                test_data.add(sample);
                test_movements_sequence.add(movement);
                test_mov_window_sequence.add(w);
            }
        }

        if(!testPhase)
            if(training_data.get(movement).isEmpty())
                training_data.remove(movement);
    }
    
    private static float min(float[] input){
        float min = 0;        
        for (int i=0; i<input.length; i++){
            min = Math.min(min, input[i]);
        }        
        return min;
    }
    
    private static float max(float[] input){
        float max = 0;        
        for (int i=0; i<input.length; i++){
            max = Math.max(max, input[i]);
        }        
        return max;
    }

    private static float mean(float[] input){
        float sumatori = 0;
        for(int i=0; i<input.length; i++)
            sumatori+= input[i];
        return (sumatori/input.length);
    }
    
    private static float std_dev(float[] input){        
        float mean = mean(input);
        float sumatori_pow2 = 0;
        for(int i=0; i<input.length; i++){
            sumatori_pow2 += input[i]*input[i];
        }
        sumatori_pow2 /= input.length;
        return (float)Math.sqrt(sumatori_pow2-(mean*mean));
    }
    
    private static float energy(float[] m){
        Complex[] m_c = new Complex[m.length];
        for(int i=0; i<m.length; i++)
            m_c[i]= new Complex(m[i],0);
        double[] tmp = FFT.magnitude(FFT.fft(m_c));

        float sumatori=0;

        //no es te en compte la frequencia 0:
        for(int i=1; i<(tmp.length/2+1); i++)
            sumatori+= tmp[i]*tmp[i];
        return sumatori/(tmp.length/2+1);
    }
    
    //root mean square
    private static float RMS(float[] input){
        float sum = 0;
        for (int i = 0; i < input.length; i++){
            sum += input[i]*input[i];
        }        
        return (float)Math.sqrt(sum/input.length);
    }
    
    private static float correlation(float[] a, float[] b){
        float mean_a = mean(a);
        float mean_b = mean(b);        
        
        float sum = 0, sum_a = 0, sum_b = 0;
        
        for(int i=0; i<a.length; i++){
            sum += (a[i]-mean_a)*(b[i]-mean_b);
            sum_a += (a[i]-mean_a)*(a[i]-mean_a);
            sum_b += (b[i]-mean_b)*(b[i]-mean_b);
        }
        
        return (float)(sum/Math.sqrt(sum_a*sum_b));
    }
    
    //zero cross rate
    private static float ZCR(float[] input){
        float sum = 0;
        for (int i=0; i<input.length-1; i++){
            if (input[i]*input[i+1] < 0)
                sum += 1;
        }
        return sum/input.length;
    }
    
    //interquartile range
    private static float IQR(float[] input){
        float iqr = 0;

        List<Float> ordered_input_list = new ArrayList<Float>();
        for (int i = 0; i < input.length; i++) {
            ordered_input_list.add(input[i]);
        }
        Collections.sort(ordered_input_list);

        float[] ordered_input = new float[input.length];
        for (int i = 0; i < input.length; i++) {
            ordered_input[i] = ordered_input_list.get(i);
        }

        int lower_quartile_index = (input.length + 1)/4;
        int upper_quartile_index = 3*(input.length + 1)/4;
        iqr = ordered_input[upper_quartile_index]-ordered_input[lower_quartile_index];

        return iqr;
    }
    
    private static float FFTentropy(float[] m){
        double[] tmp = half_fftModule(m);
        return 0;
    }

    private static double[] half_fftModule(float[] m){
        Complex[] m_c = new Complex[m.length];
        for(int i=0; i<m.length; i++)
            m_c[i]= new Complex(m[i],0);
        double[] tmp = FFT.magnitude(FFT.fft(m_c));
        double[] result = new double[tmp.length/2];
        for(int i=0; i<result.length; i++)
            result[i] = tmp[i+1]; //+1 per treure la frequencia 0.
        return result;
    }
    
    private static double[] join_frequencies(double[] vector){
        if(vector.length%2!=0)
            return null;

        int m = 4;
        switch(vector.length){
            case 64:
                m = 8; break;
            case 32:
                m = 4; break;
            case 16:
                m = 2; break;
        }
        int n = vector.length / m;

        double[] vector_band = new double[vector.length/m];
        for(int i=0; i<n; i++){
            float sumatori = 0;
            for(int j=0; j<m; j++)
                sumatori+=vector[i*m+j];
            vector_band[i]=sumatori/m;
        }
        return vector_band;
    } 
    
    private static boolean isNull(float[] m_w){
        int counter = 0;
        float[] m_fourth = new float[m_w.length/4];
        for(int j=0; j<4; j++){
            for(int i=0; i<m_w.length/4; i++)
                m_fourth[i] = m_w[j*m_w.length/4+i];
            if(energy(m_fourth)<1)
                counter++;
        }
        if(counter>2)
            return true;
        else
            return false;
    }
    
    //statistical features for the acceleration signals:
    //energy, min, max, mean, standard deviation, root mean squared acceleration
    //interquartile range and zero crossings for each signal
    //and correlation, FFT
    private static Point sample(float[] x_w, float[] y_w, float[] z_w, float[] m_w, int w){
        float energy;
        double[] xFFT, yFFT, zFFT, mFFT, xFFT_join, yFFT_join, zFFT_join, mFFT_join;

        energy = energy(m_w);
        if( energy < Cons.min_energy_window )
            return null;

        for(int i=0; i<window; i++){
            x_w[i] /= (float) Math.sqrt(energy);
            y_w[i] /= (float) Math.sqrt(energy);
            z_w[i] /= (float) Math.sqrt(energy);
            m_w[i] = (float) Math.sqrt(x_w[i]*x_w[i]+y_w[i]*y_w[i]+z_w[i]*z_w[i]);
        }
        
        List<Float> features = new ArrayList<Float>();
//        
//        features.add(mean(x_w));
//        features.add(mean(y_w));
//        features.add(mean(z_w));
//        features.add(mean(m_w));
//        
//        features.add(std_dev(x_w));
//        features.add(std_dev(y_w));
//        features.add(std_dev(z_w));
//        features.add(std_dev(m_w));
//        
//        features.add(min(x_w));
//        features.add(max(x_w));
//        features.add(min(y_w));
//        features.add(max(y_w));
//        features.add(min(z_w));
//        features.add(max(z_w));
//        features.add(min(m_w));
//        features.add(max(m_w));
//        
        features.add(energy(x_w));
        features.add(energy(y_w));
        features.add(energy(z_w));
        features.add(energy(m_w));
        
//        features.add(correlation(x_w, y_w));
//        features.add(correlation(y_w, z_w));
//        features.add(correlation(z_w, x_w));
//
//        features.add(RMS(x_w));
//        features.add(ZCR(x_w));
//        features.add(IQR(x_w));
//
//        features.add(RMS(y_w));
//        features.add(ZCR(y_w));
//        features.add(IQR(y_w));
//
//        features.add(RMS(z_w));
//        features.add(ZCR(z_w));
//        features.add(IQR(z_w));
//
//        features.add(RMS(m_w));
//        features.add(ZCR(m_w));
//        features.add(IQR(m_w));
//
        xFFT = half_fftModule(x_w);
        yFFT = half_fftModule(y_w);
        zFFT = half_fftModule(z_w);
        mFFT = half_fftModule(m_w);

        xFFT_join = join_frequencies(xFFT);
        yFFT_join = join_frequencies(yFFT);
        zFFT_join = join_frequencies(zFFT);
        mFFT_join = join_frequencies(mFFT);        

        for(int i=0; i<mFFT_join.length/1; i++)
            features.add((float)mFFT_join[i]);
        for(int i=0; i<xFFT_join.length/1; i++)
            features.add((float)xFFT_join[i]);
        for(int i=0; i<yFFT_join.length/1; i++)
            features.add((float)yFFT_join[i]);
        for(int i=0; i<zFFT_join.length/1; i++)
            features.add((float)zFFT_join[i]);
              
        return new Point(features);
    }
//    
//    
//    public static void testFilter(String dir, String fileName){
//        BufferedReader file=null;
//        PrintWriter wfile1 = null;
//        PrintWriter wfile2 = null;
//        
//        try{            
//            file = new BufferedReader(new FileReader(dir+fileName+".txt"));
//            
//            wfile1 = new PrintWriter(new FileWriter(dir+"separated/"+fileName+"_x.txt"));
//            wfile2 = new PrintWriter(new FileWriter(dir+"separated/"+fileName+"_xf.txt"));
//            //load data from file
//            String line;
//            List<Float> listX = new ArrayList<Float>();
//            List<Float> listY = new ArrayList<Float>();
//            List<Float> listZ = new ArrayList<Float>();
//            
//            int size = 0;
//            while((line = file.readLine())!=null){
//                size++;
//                String[] figures = line.split(" ");
//                listX.add(new Float(figures[1]));
//                listY.add(new Float(figures[2]));
//                listZ.add(new Float(figures[3]));
//            }
//            
//            float[] x_ = new float[size];
////            float[] y_ = new float[size];
////            float[] z_ = new float[size];
////            float[] m_ = new float[size];
//            
//            for (int i=0; i<size; i++){
//                x_[i] = listX.get(i);
////                y_[i] = listY.get(i);
////                //remove the force of gravity
////                z_[i] = listZ.get(i) - (float)9.8;
////                m_[i] = (float) Math.sqrt(x_[i]*x_[i]+y_[i]*y_[i]+z_[i]*z_[i]);
//            }
//            
//            KalmanFilter kf = new KalmanFilter(5, 15);
//            kf.setInitState(x_[0], 0.1);
//            
//            for (int i=0; i<size; i++){
//                kf.correct(x_[i]);
//                wfile1.println(listX.get(i));
//                wfile2.println(kf.getState());
//            }
//            
//        } catch (Exception ex) {
//            ex.printStackTrace();
//        }
//        try {
//            if(file!=null)
//                file.close();
//            if(wfile1!=null)
//                wfile1.close();
//            if(wfile2!=null)
//                wfile2.close();
//        } catch (IOException ex1) {
//            ex1.printStackTrace();
//        }
//    }
}
