package util;


import java.util.ArrayList;
import java.util.List;

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

/**
 *
 * @author Dunst
 */
public class Cluster {
    private int id;
    private Point center;
    private List<Point> samples;
    
    public Cluster(int id, Point center){
        this.id = id;
        this.center = center.copy();
        samples = new ArrayList<Point>();
    }
    
    public Cluster(int id, List<Point> points){
        this.id = id;
        this.samples.addAll(points);
        this.center = calculateCenter(points);
    }
    
    public Cluster(int id){
        this.id = id;
        samples = new ArrayList<Point>();
    }
    
    public void add(Point sample){
        this.samples.add(sample);
    }
    
    public void addAll(List<Point> samples){
        this.samples.addAll(samples);
    }
    
    public void clear(){
        this.samples.clear();
    }
        
    public double euclidDistance(Point point){
        return Util.euclidDistance(point, getCenter());
    }
    
    public int size(){
        return this.samples.size();
    }

    /**
     * @return the id
     */
    public int getId() {
        return id;
    }

    /**
     * @param id the id to set
     */
    public void setId(int id) {
        this.id = id;
    }

    /**
     * @return the center
     */
    public Point getCenter() {
        return center.copy();
    }
    
    public void setCenter(Point center) {       
        this.center = center.copy();
    }

   private Point calculateCenter(List<Point> points){
        Point centroid = new Point();
        
        for (Point sample: points){
            centroid = Util.addPoints(centroid, sample);
        }
        
        centroid = Util.divideN(centroid, points.size());
        
        return centroid;
    }
   
   private Point calculateCenterPerFeature(){
        Point centroid = new Point();
        
        for (Point sample: this.samples){
            centroid = Util.addPoints(centroid, sample);
        }
        
        centroid = Util.divideN(centroid, this.samples.size());
        
        return centroid;
    }
   
   public void resetCenter(){
       this.center = calculateCenterPerFeature();
   }
   
   public List<Point> getSamples(){
       return this.samples;
   }
}
