/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package util;

import Jama.Matrix;
import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Dunst
 */
public class Util {
    public static boolean equals(Point a, Point b){
        if (a == null || b == null)
            return false;
        if (a.size() != b.size())
            return false;
        for(int i = 0; i < a.size(); i++){
            if (a.get(i) != b.get(i))
                return false;
        } 
        return true;
    }
    
    public static boolean belongs(Point element, List<Point> data){
        for(int i = 0; i < data.size(); i++){
            if (equals(data.get(i), element)){
                return true;
            }
        }
        return false;
    }
    
    public static double euclidDistance(Point a, Point b){
        double d = 0;
        
        for (int i=0; i< a.size(); i++){
            d += Math.pow((a.get(i)-b.get(i)),2);
        }
        if (d != 0)
            d = Math.sqrt(d);
        
        return d;
    }
    
//    public static double getMultivariateGaussian(Point sample, Point mean, Matrix covariance){
//        double multivariateGaussian = Math.pow(Math.pow(2.0*Math.PI, sample.size()) *covariance.det(), -0.5);
//
//        Matrix x = new Matrix(1, sample.size());
//        Matrix xt = new Matrix(sample.size(), 1);
//        
//        for(int i = 0; i < sample.size(); i++){
//            xt.set(i, 0, sample.get(i)-mean.get(i));
//            x.set(0, i, -xt.get(i, 0)/2);
//        }
//        Matrix s1 = covariance.inverse();
//        Matrix epow = s1.times(xt);
//        epow = x.times(epow);
//        
//        multivariateGaussian *= Math.exp(epow.det());
//
//        return multivariateGaussian;
//    }
    
    //assumes that the covariance matrix is diagonal
    public static double getMultivariateGaussian(Point sample, Point mean, double[] stdDevs){
        double gaussian = 1;
        
        for(int i = 0; i < sample.size(); i++){
            gaussian *= Util.getUnivariateGaussian(sample.get(i), mean.get(i), stdDevs[i]);
        }
        return gaussian;
    }
    
    private static double getUnivariateGaussian(double x, double mean, double stdDev){
        return Math.exp(-0.5 * Math.pow((x - mean)/stdDev, 2))/(stdDev * Cons.sqrt2PI);
    }
    
    public static Matrix getCovarianceMatrix(List<Point> samples, Point mean){
        int numOfFeatures = mean.size();
        Matrix covMatrix = new Matrix(numOfFeatures, numOfFeatures);
        for(int i = 0; i < numOfFeatures; i++){
                for(int j = 0; j <= i; j++){
                    double covariance = 0;
                    for(int k = 0; k < samples.size(); k++){
                        covariance += (samples.get(k).get(i) - mean.get(i))*(samples.get(k).get(j) - mean.get(j));
                    }
                    covMatrix.set(i, j, covariance/samples.size());
                    covMatrix.set(j, i, covMatrix.get(i, j));
                }
        }

        return covMatrix;
    }
    
    public static Point addPoints(Point a, Point b){
        List<Float> _params = new ArrayList<Float>();
        if (a.size() != b.size()){
            return b.copy();
        }
        else{   
            for (int i=0; i<b.size(); i++){
                _params.add(i, a.get(i)+b.get(i));         
            }
        }
        return new Point(_params);
    }
    
    public static Point times(Point a, float d){
        List<Float> _params = new ArrayList<Float>();
        for (int i=0; i<a.size(); i++){
            _params.add(i, a.get(i)*d);
        }
        return new Point(_params);
    }
    
    public static Point divideN(Point a, float d){
        List<Float> _params = new ArrayList<Float>();
        for (int i=0; i<a.size(); i++){
            _params.add(i, a.get(i)/d);
        }
        return new Point(_params);
    }
    
//    public static List<Double> getVariance(List<Point> samples, Point mean){
//        List<Double> variances = new ArrayList<Double>();
//        double[] varianceValues = new double[mean.size()];
//        for(int i = 0; i < varianceValues.length; i++)
//            varianceValues[i] = 0;
//        for(int i = 0; i < varianceValues.length; i++)
//            for(int j = 0; j < samples.size(); j++){
//                Point sample = samples.get(j);
//                varianceValues[i] += Math.pow(sample.get(i) - mean.get(i), 2);
//            }
//        for(int i = 0; i < varianceValues.length; i++)
//            variances.add(varianceValues[i]/samples.size());
//        
//        return variances;        
//    }
    
    public static double[] getVariance(List<Point> samples, Point mean){
        int numOfFeatures = mean.size();
        double[] variance = new double[numOfFeatures];
        for(int i = 0; i < numOfFeatures; i++)
            variance[i] = 0;
        for(int i = 0; i < numOfFeatures; i++){
            for(int j = 0; j < samples.size(); j++){
                variance[i] += Math.pow(samples.get(j).get(i) - mean.get(i), 2);
            }            
        }
        for(int i = 0; i < numOfFeatures; i++)
            variance[i] /= samples.size();
        
        return variance;
    }
}
